/*
 * SonarQube Java
 * Copyright (C) 2012-2021 SonarSource SA
 * mailto:info AT sonarsource DOT com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package org.sonar.java.regex.ast;

import java.util.List;
import org.junit.jupiter.api.Test;

import static org.assertj.core.api.Assertions.assertThat;
import static org.sonar.java.regex.RegexParserTestUtils.assertKind;
import static org.sonar.java.regex.RegexParserTestUtils.assertSingleEdge;
import static org.sonar.java.regex.RegexParserTestUtils.assertSuccessfulParse;
import static org.sonar.java.regex.RegexParserTestUtils.assertType;

class CapturingGroupTreeTest {

  @Test
  void numbering() {
    RegexTree tree = assertSuccessfulParse("((A)(?:N)(B(?<groupC>C)))");
    assertKind(RegexTree.Kind.CAPTURING_GROUP, tree);
    CapturingGroupTree abc = ((CapturingGroupTree) tree);
    assertThat(abc.getGroupNumber()).isEqualTo(1);
    assertThat(abc.getName()).isEmpty();

    RegexTree abcElement = abc.getElement();
    assertKind(RegexTree.Kind.SEQUENCE, abcElement);
    List<RegexTree> abcItems = ((SequenceTree) abcElement).getItems();
    assertThat(abcItems).hasSize(3);
    assertThat(abcItems.stream().map(RegexTree::kind)).containsExactly(RegexTree.Kind.CAPTURING_GROUP, RegexTree.Kind.NON_CAPTURING_GROUP, RegexTree.Kind.CAPTURING_GROUP);

    CapturingGroupTree a = ((CapturingGroupTree) abcItems.get(0));
    CapturingGroupTree bc = ((CapturingGroupTree) abcItems.get(2));

    assertThat(a.getGroupNumber()).isEqualTo(2);
    assertThat(a.getName()).isEmpty();

    assertThat(bc.getGroupNumber()).isEqualTo(3);
    assertThat(bc.getName()).isEmpty();

    RegexTree bcElement = bc.getElement();
    assertKind(RegexTree.Kind.SEQUENCE, bcElement);
    List<RegexTree> bcItems = ((SequenceTree) bcElement).getItems();
    assertThat(bcItems).hasSize(2);

    assertKind(RegexTree.Kind.CHARACTER, bcItems.get(0));
    assertKind(RegexTree.Kind.CAPTURING_GROUP, bcItems.get(1));

    CapturingGroupTree c = ((CapturingGroupTree) bcItems.get(1));
    assertThat(c.getGroupNumber()).isEqualTo(4);
    assertThat(c.getName()).hasValue("groupC");

    testAutomaton(abc, abcItems, a, bc, bcItems, c);
  }

  private void testAutomaton(CapturingGroupTree abc, List<RegexTree> abcItems, CapturingGroupTree a, CapturingGroupTree bc, List<RegexTree> bcItems, CapturingGroupTree c) {
    assertThat(abc.incomingTransitionType()).isEqualTo(AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(abc, abc.getElement(), AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(abc.getElement(), a, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(a, a.getElement(), AutomatonState.TransitionType.CHARACTER);
    NonCapturingGroupTree n = assertType(NonCapturingGroupTree.class, abcItems.get(1));
    RegexTree nElement = n.getElement();
    assertThat(nElement).isNotNull();
    EndOfCapturingGroupState endOfA = getAndAssertEndOfCapturingGroup(a);
    assertSingleEdge(a.getElement(), endOfA, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(endOfA, n, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(n, nElement, AutomatonState.TransitionType.CHARACTER);
    assertSingleEdge(nElement, bc, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(bc, bc.getElement(), AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(bc.getElement(), bcItems.get(0), AutomatonState.TransitionType.CHARACTER);
    assertSingleEdge(bcItems.get(0), c, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(c, c.getElement(), AutomatonState.TransitionType.CHARACTER);
    EndOfCapturingGroupState endOfC = getAndAssertEndOfCapturingGroup(c);
    EndOfCapturingGroupState endOfBC = getAndAssertEndOfCapturingGroup(bc);
    EndOfCapturingGroupState endOfABC = getAndAssertEndOfCapturingGroup(abc);
    assertSingleEdge(c.getElement(), endOfC, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(endOfC, endOfBC, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(endOfBC, endOfABC, AutomatonState.TransitionType.EPSILON);
    assertSingleEdge(endOfABC, abc.continuation(), AutomatonState.TransitionType.EPSILON);
  }

  private EndOfCapturingGroupState getAndAssertEndOfCapturingGroup(CapturingGroupTree group) {
    EndOfCapturingGroupState endOfGroup = assertType(EndOfCapturingGroupState.class, group.getElement().continuation());
    assertThat(endOfGroup.group()).isSameAs(group);
    return endOfGroup;
  }

}
